/*
 * This file is part of PowerDNS or dnsdist.
 * Copyright -- PowerDNS.COM B.V. and its contributors
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * In addition, for the avoidance of any doubt, permission is granted to
 * link this program with OpenSSL and to (re)distribute the binaries
 * produced as the result of such linking.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */
#pragma once

class ExponentialBackOffTimer
{
public:
  ExponentialBackOffTimer(unsigned int maxBackOff) :
    d_maxBackOff(maxBackOff)
  {
  }

  unsigned int get(size_t consecutiveFailures) const
  {
    unsigned int backOff = d_maxBackOff;
    if (consecutiveFailures <= 31) {
      backOff = 1U << consecutiveFailures;
      backOff = std::min(d_maxBackOff, backOff);
    }
    return backOff;
  }

private:
  const unsigned int d_maxBackOff;
};
