// !! This file has been generated by dnsdist-rules-generator.py, do not edit by hand!!
luaCtx.writeFunction("AllowResponseAction", []() {
  return dnsdist::actions::getAllowResponseAction();
});
luaCtx.writeFunction("DelayResponseAction", [](uint32_t msec) {
  return dnsdist::actions::getDelayResponseAction(msec);
});
luaCtx.writeFunction("DropResponseAction", []() {
  return dnsdist::actions::getDropResponseAction();
});
luaCtx.writeFunction("LogResponseAction", [](std::optional<std::string> file_name, std::optional<bool> append, std::optional<bool> buffered, std::optional<bool> verbose_only, std::optional<bool> include_timestamp) {
  return dnsdist::actions::getLogResponseAction(file_name ? *file_name : "", append ? *append : false, buffered ? *buffered : false, verbose_only ? *verbose_only : true, include_timestamp ? *include_timestamp : false);
});
luaCtx.writeFunction("LuaFFIPerThreadResponseAction", [](std::string code) {
  return dnsdist::actions::getLuaFFIPerThreadResponseAction(code);
});
luaCtx.writeFunction("SetEDNSOptionResponseAction", [](uint32_t code, std::string data) {
  return dnsdist::actions::getSetEDNSOptionResponseAction(code, data);
});
luaCtx.writeFunction("SetExtendedDNSErrorResponseAction", [](uint16_t info_code, std::optional<std::string> extra_text, std::optional<bool> clear_existing_entries) {
  return dnsdist::actions::getSetExtendedDNSErrorResponseAction(info_code, extra_text ? *extra_text : "", clear_existing_entries);
});
luaCtx.writeFunction("SetReducedTTLResponseAction", [](uint8_t percentage) {
  return dnsdist::actions::getSetReducedTTLResponseAction(percentage);
});
luaCtx.writeFunction("SetSkipCacheResponseAction", []() {
  return dnsdist::actions::getSetSkipCacheResponseAction();
});
luaCtx.writeFunction("SetTagResponseAction", [](std::string tag, std::string value) {
  return dnsdist::actions::getSetTagResponseAction(tag, value);
});
luaCtx.writeFunction("SNMPTrapResponseAction", [](std::optional<std::string> reason) {
  return dnsdist::actions::getSNMPTrapResponseAction(reason ? *reason : "");
});
luaCtx.writeFunction("TCResponseAction", []() {
  return dnsdist::actions::getTCResponseAction();
});
luaCtx.writeFunction("UnsetTagResponseAction", [](std::string tag) {
  return dnsdist::actions::getUnsetTagResponseAction(tag);
});
